// DXSound.cpp: implementation of the CDXSound class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DXMisc.h"
#include "DXSound.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define ER_MEM					0xe000
#define ER_CANNOTOPEN 			0xe100
#define ER_NOTWAVEFILE 			0xe101
#define ER_CANNOTREAD 			0xe102
#define ER_CORRUPTWAVEFILE		0xe103
#define ER_CANNOTWRITE			0xe104

int WaveOpenFile (LPCSTR pszFileName, HMMIO *phmmioIn, WAVEFORMATEX **ppwfxInfo,
				 MMCKINFO *pckInRIFF, BOOL bResource = FALSE);
int WaveStartDataRead (HMMIO *, MMCKINFO *, MMCKINFO *);
int WaveReadFile (HMMIO, UINT, BYTE *, MMCKINFO *, UINT *);
void WaveCloseReadFile (HMMIO *, WAVEFORMATEX **);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC (CDXSound, CObject);

CDXSound::CDXSound ()
{
	m_pDS = NULL;
	m_pDSBStreamBuffer = NULL;
	m_pWnd = NULL;
	memset (&m_wiWave, 0, sizeof (WAVEINFOCA));
}

CDXSound::~CDXSound ()
{
	Destroy ();
}

BOOL CDXSound::Load (CWnd* pWnd, UINT idFilename)
{
	return LoadHelper (pWnd, MAKEINTRESOURCE (idFilename), TRUE);
}

BOOL CDXSound::Load (CWnd* pWnd, LPCSTR szFilename)
{
	return LoadHelper (pWnd, szFilename, FALSE);
}

BOOL CDXSound::LoadHelper (CWnd* pWnd, LPCSTR szFilename, BOOL bResource)
{
	Destroy ();

	int nChkErr = WaveOpenFile (szFilename,
		&m_wiWave.hmmio, &m_wiWave.pwfx, &m_wiWave.mmckInRIFF, bResource);
	if (nChkErr != 0 || m_wiWave.pwfx->wFormatTag != WAVE_FORMAT_PCM) {
		Destroy ();
		return FALSE;
	}

	// Seek to the data chunk. mmck.ckSize will be the size of all the data in the file.
	nChkErr = WaveStartDataRead (&m_wiWave.hmmio, &m_wiWave.mmck, &m_wiWave.mmckInRIFF);
	if (nChkErr != 0) {
		Destroy ();
		return FALSE;
	}
	
	//   DirectSound
	HRESULT dsrval = DirectSoundCreate (NULL, &m_pDS, NULL);
	if (dsrval != DS_OK || m_pDS == NULL) {
		debugDSReportError ("CDXSound::Load:DirectSoundCreate", dsrval);
		Destroy ();
		return FALSE;
	}

	//  
	dsrval = m_pDS->SetCooperativeLevel (pWnd->GetSafeHwnd (), DSSCL_NORMAL);
	if (dsrval != DS_OK) {
		debugDSReportError ("CDXSound::Load:SetCooperativeLevel", dsrval);
		Destroy ();
		return FALSE;
	}

	m_pWnd = pWnd;

	 // Create the secondary DirectSoundBuffer object to receive our sound data.
	DSBUFFERDESC dsbd;
	memset (&dsbd, 0, sizeof (DSBUFFERDESC));
	dsbd.dwSize = sizeof (DSBUFFERDESC);
	dsbd.dwFlags = /* DSBCAPS_CTRLDEFAULT |*/ DSBCAPS_STATIC |
		DSBCAPS_GLOBALFOCUS | DSBCAPS_STICKYFOCUS;
	dsbd.dwBufferBytes = m_wiWave.mmck.cksize;
	dsbd.lpwfxFormat = m_wiWave.pwfx;
	dsrval = m_pDS->CreateSoundBuffer (&dsbd, &m_pDSBStreamBuffer, NULL);						
	if (dsrval != DS_OK || m_pDSBStreamBuffer == NULL) {
		debugDSReportError ("CDXSound::Load:CreateSoundBuffer", dsrval);
		Destroy ();
		return FALSE;
	}

    LPBYTE lpWrite1, lpWrite2;
    DWORD dwLen1, dwLen2;
    UINT uActualBytesWritten;

    dsrval = m_pDSBStreamBuffer->Lock (0, m_wiWave.mmck.cksize, (LPVOID*) &lpWrite1, &dwLen1,
		(LPVOID*) &lpWrite2, &dwLen2, 0);
    if (dsrval != DS_OK) {
		debugDSReportError ("CDXSound::Load:Lock", dsrval);
		Destroy ();
		return FALSE;
	}

    WaveReadFile (m_wiWave.hmmio, (UINT) dwLen1, lpWrite1,
		&m_wiWave.mmck, &uActualBytesWritten);
	if (nChkErr != 0) {
		m_pDSBStreamBuffer->Unlock ((LPVOID) lpWrite1, dwLen1, NULL, 0);
		Destroy ();
		return FALSE;
	}

	m_pDSBStreamBuffer->Unlock ((LPVOID) lpWrite1, dwLen1, NULL, 0);

	WaveCloseReadFile (&m_wiWave.hmmio, &m_wiWave.pwfx);

	return TRUE;
}

void CDXSound::Destroy (void)
{
	Stop ();

	WaveCloseReadFile (&m_wiWave.hmmio, &m_wiWave.pwfx);

	if (m_pDSBStreamBuffer != NULL) {
		m_pDSBStreamBuffer->Release ();
		m_pDSBStreamBuffer = NULL;
	}

	m_pWnd = NULL;

	if (m_pDS != NULL) {
		m_pDS->Release ();
		m_pDS = NULL;
	}
}

BOOL CDXSound::Rewind (void)
{
	if (m_pDSBStreamBuffer == NULL)
		return FALSE;

	HRESULT dsrval = m_pDSBStreamBuffer->SetCurrentPosition (0);
	if (dsrval != DS_OK) {
		debugDSReportError ("CDXSound::Rewind:SetCurrentPosition", dsrval);
		return FALSE;
	}

	return TRUE;
}

BOOL CDXSound::Play (BOOL Loop)
{
	Stop ();

	if (m_pDSBStreamBuffer == NULL)
		return FALSE;

	HRESULT dsrval = m_pDSBStreamBuffer->Play (0, 0, Loop ? DSBPLAY_LOOPING : 0);
	if (dsrval != DS_OK) {
		debugDSReportError ("CDXSound::Play:Play", dsrval);
		return FALSE;
	}

	return TRUE;
}

BOOL CDXSound::Stop (void)
{
	if (m_pDSBStreamBuffer == NULL)
		return FALSE;

	m_pDSBStreamBuffer->Stop ();
	while (IsPlaying ())
		Sleep (0);

	return TRUE;
}

BOOL CDXSound::IsPlaying (void)
{
	if (m_pDSBStreamBuffer == NULL)
		return FALSE;

	DWORD dwStatus = 0;
	HRESULT dsrval = m_pDSBStreamBuffer->GetStatus (&dwStatus);
	return ((dsrval == DS_OK) && (dwStatus & DSBSTATUS_PLAYING));
}

BOOL CDXSound::IsLost (void)
{
	if (m_pDSBStreamBuffer == NULL)
		return FALSE;

	DWORD dwStatus = 0;
	HRESULT dsrval = m_pDSBStreamBuffer->GetStatus (&dwStatus);
	return ((dsrval == DS_OK) && (dwStatus & DSBSTATUS_BUFFERLOST));
}

//	This function will open a wave input file and prepare it for reading, so the data can
//	be easily read with WaveReadFile. Returns 0 if successful, the error code if not.
//	pszFileName	- Input filename to load or resource id
//	phmmioIn	- Pointer to handle which will be used for further mmio routines
//	ppwfxInfo	- Ptr to ptr to WaveFormatEx structure with all info about the file
//	bResource	- resource or disk file
int WaveOpenFile (LPCSTR pszFileName, HMMIO *phmmioIn, WAVEFORMATEX **ppwfxInfo,
				 MMCKINFO *pckInRIFF, BOOL bResource)
{
	HMMIO           hmmioIn = NULL;
	MMCKINFO        ckIn;			// chunk info. for general use.
	PCMWAVEFORMAT   pcmWaveFormat;	// Temp PCM structure to load in.
	WORD            cbExtraAlloc;	// Extra bytes for waveformatex
	int             nError = 0;		// Return value.

	*ppwfxInfo = NULL;
	
	if (bResource) {
		//   
		MMIOINFO mmio;
		memset (&mmio, 0, sizeof (MMIOINFO));
		mmio.fccIOProc = FOURCC_MEM;

		HRSRC hRC = FindResource (NULL, pszFileName, "WAVE");
		if (hRC == NULL) {
			nError = ER_CANNOTOPEN;
			ASSERT (FALSE);
			goto ERROR_READING_WAVE;
		}
		mmio.cchBuffer = SizeofResource (NULL, hRC);

		HGLOBAL hGlobal = LoadResource (NULL, hRC);
		if (hGlobal == NULL) {
			nError = ER_CANNOTOPEN;
			ASSERT (FALSE);
			goto ERROR_READING_WAVE;
		}

		mmio.pchBuffer = (HPSTR) LockResource (hGlobal);
		if (mmio.pchBuffer == NULL) {
			nError = ER_CANNOTOPEN;
			ASSERT (FALSE);
			goto ERROR_READING_WAVE;
		}

		if ((hmmioIn = mmioOpen (NULL, &mmio,
			MMIO_ALLOCBUF | MMIO_READ | MMIO_DENYWRITE)) == NULL) {
			nError = ER_CANNOTOPEN;
			goto ERROR_READING_WAVE;
		}
	} else {
		//  
		if ((hmmioIn = mmioOpen ((LPSTR) pszFileName, NULL,
			MMIO_ALLOCBUF | MMIO_READ | MMIO_DENYWRITE)) == NULL) {
			nError = ER_CANNOTOPEN;
			goto ERROR_READING_WAVE;
		}
	}

	if ((nError = (int)mmioDescend (hmmioIn, pckInRIFF, NULL, 0)) != 0) {
		goto ERROR_READING_WAVE;
	}

	if ((pckInRIFF->ckid != FOURCC_RIFF) ||
		(pckInRIFF->fccType != mmioFOURCC ('W', 'A', 'V', 'E'))) {
		nError = ER_NOTWAVEFILE;
		goto ERROR_READING_WAVE;
	}
			
	// Search the input file for for the 'fmt ' chunk.
    ckIn.ckid = mmioFOURCC ('f', 'm', 't', ' ');
    if ((nError = (int) mmioDescend(hmmioIn, &ckIn, pckInRIFF, MMIO_FINDCHUNK)) != 0) {
		goto ERROR_READING_WAVE;
	}
					
	/* Expect the 'fmt' chunk to be at least as large as <PCMWAVEFORMAT>;
    * if there are extra parameters at the end, we'll ignore them */
    if (ckIn.cksize < (long) sizeof(PCMWAVEFORMAT)) {
		nError = ER_NOTWAVEFILE;
		goto ERROR_READING_WAVE;
	}
															
	/* Read the 'fmt ' chunk into <pcmWaveFormat>.*/
    if (mmioRead(hmmioIn, (HPSTR) &pcmWaveFormat, (long) sizeof(pcmWaveFormat))
		!= (long) sizeof(pcmWaveFormat)) {
		nError = ER_CANNOTREAD;
		goto ERROR_READING_WAVE;
	}							

	// Ok, allocate the waveformatex, but if its not pcm
	// format, read the next word, and thats how many extra
	// bytes to allocate.
	if (pcmWaveFormat.wf.wFormatTag == WAVE_FORMAT_PCM)
		cbExtraAlloc = 0;
	else {
		// Read in length of extra bytes.
		if (mmioRead (hmmioIn, (LPSTR) &cbExtraAlloc,
			(long) sizeof(cbExtraAlloc)) != (long) sizeof (cbExtraAlloc)) {
			nError = ER_CANNOTREAD;
			goto ERROR_READING_WAVE;
		}

	}
							
	// Ok, now allocate that waveformatex structure.
	*ppwfxInfo = (WAVEFORMATEX*) GlobalAlloc (GMEM_FIXED, sizeof (WAVEFORMATEX) +
		cbExtraAlloc);
	if (*ppwfxInfo== NULL) {
		nError = ER_MEM;
		goto ERROR_READING_WAVE;
	}

	// Copy the bytes from the pcm structure to the waveformatex structure
	memcpy (*ppwfxInfo, &pcmWaveFormat, sizeof(pcmWaveFormat));
	(*ppwfxInfo)->cbSize = cbExtraAlloc;

	// Now, read those extra bytes into the structure, if cbExtraAlloc != 0.
	if (cbExtraAlloc != 0) {
		if (mmioRead(hmmioIn, (LPSTR) (((BYTE*)&((*ppwfxInfo)->cbSize))+sizeof(cbExtraAlloc)),
			(long) (cbExtraAlloc)) != (long) (cbExtraAlloc)) {
			nError = ER_NOTWAVEFILE;
			goto ERROR_READING_WAVE;
		}
	}

	/* Ascend the input file out of the 'fmt ' chunk. */
	if ((nError = mmioAscend(hmmioIn, &ckIn, 0)) != 0)
		goto ERROR_READING_WAVE;

	goto TEMPCLEANUP;

ERROR_READING_WAVE:
	if (*ppwfxInfo != NULL) {
		GlobalFree(*ppwfxInfo);
		*ppwfxInfo = NULL;
	}
	if (hmmioIn != NULL) {
		mmioClose(hmmioIn, 0);
		hmmioIn = NULL;
	}
	
TEMPCLEANUP:
	*phmmioIn = hmmioIn;

	return (nError);

}

//	This routine has to be called before WaveReadFile as it searchs for the chunk
//	to descend into for	reading, that is, the 'data' chunk.  For simplicity, this
//	used to be in the open routine, but was	taken out and moved to a separate
//	routine so there was more control on the chunks that are before the data chunk,
//	such as 'fact', etc...
int WaveStartDataRead (HMMIO *phmmioIn, MMCKINFO *pckIn, MMCKINFO *pckInRIFF)
{
	int nError = mmioSeek (*phmmioIn, pckInRIFF->dwDataOffset + sizeof(FOURCC), SEEK_SET);
	
	// Do a nice little seek...
	if (nError == -1) {
		goto ERROR_READING_WAVE;
	}

	// Search the input file for for the 'data' chunk.
	pckIn->ckid = mmioFOURCC ('d', 'a', 't', 'a');
	nError = mmioDescend (*phmmioIn, pckIn, pckInRIFF, MMIO_FINDCHUNK);
	if (nError != 0) {
		goto ERROR_READING_WAVE;
	}

	goto CLEANUP;

ERROR_READING_WAVE:

CLEANUP:
	return(nError);
}

//	This will read wave data from the wave file. Makre sure we're descended into
//	the data chunk, else this will fail bigtime!
//	hmmioIn			- Handle to mmio
//	cbRead			- # of bytes to read
//	pbDest			- Destination buffer to put bytes
//	cbActualRead	- # of bytes actually read
int WaveReadFile (HMMIO hmmioIn, UINT cbRead, BYTE *pbDest, MMCKINFO *pckIn,
				  UINT *cbActualRead)
{
	MMIOINFO mmioinfoIn;
	UINT cT, cbDataIn, uCopyLength;

	int nError = mmioGetInfo(hmmioIn, &mmioinfoIn, 0);
	if (nError != 0) {
		goto ERROR_CANNOT_READ;
	}
				
	cbDataIn = cbRead;
	if (cbDataIn > pckIn->cksize)
		cbDataIn = pckIn->cksize;

	pckIn->cksize -= cbDataIn;	
	for (cT = 0; cT < cbDataIn;){
		/* Copy the bytes from the io to the buffer. */
		if (mmioinfoIn.pchNext == mmioinfoIn.pchEndRead) {
			if ((nError = mmioAdvance(hmmioIn, &mmioinfoIn, MMIO_READ)) != 0) {
				goto ERROR_CANNOT_READ;
			}
			if (mmioinfoIn.pchNext == mmioinfoIn.pchEndRead) {
				nError = ER_CORRUPTWAVEFILE;
				goto ERROR_CANNOT_READ;
			}
		}
		
		// Actual copy.
		uCopyLength = (UINT) (mmioinfoIn.pchEndRead - mmioinfoIn.pchNext);
		if ((cbDataIn - cT) < uCopyLength)
			uCopyLength = cbDataIn - cT;
		memcpy ((BYTE*) (pbDest+cT), (BYTE*) mmioinfoIn.pchNext, uCopyLength);
		cT += uCopyLength;
		mmioinfoIn.pchNext += uCopyLength;
		}

	if ((nError = mmioSetInfo(hmmioIn, &mmioinfoIn, 0)) != 0) {
		goto ERROR_CANNOT_READ;
	}

	*cbActualRead = cbDataIn;
	goto FINISHED_READING;

ERROR_CANNOT_READ:
	*cbActualRead = 0;

FINISHED_READING:
	return nError;
}

//	This will close the wave file openned with WaveOpenFile.
//	Returns 0 if successful, non-zero if there was a warning.
//	phmmioIn	- Pointer to the handle to input MMIO
//	ppwfxSrc	- Pointer to pointer to WaveFormatEx structure
void WaveCloseReadFile (HMMIO *phmmio, WAVEFORMATEX **ppwfxSrc)
{
	if (*ppwfxSrc != NULL) {
		GlobalFree (*ppwfxSrc);
		*ppwfxSrc = NULL;
	}

	if (*phmmio != NULL) {
		mmioClose (*phmmio, 0);
		*phmmio = NULL;
	}
}
