// DXSprite.cpp: implementation of the CDCDXSprite class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DXMisc.h"
#include "DXSprite.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC (CDXSprite, CObject);

CDXSprite::CDXSprite ()
{
	::InitializeCriticalSection (&m_cs);
	m_dwFrame = 0;
	m_dwSize = 0;
	m_x = 0;
	m_y = 0;
	m_strName = _T("");
}

CDXSprite::~CDXSprite ()
{
	Destroy ();
	::DeleteCriticalSection (&m_cs);
}

BOOL CDXSprite::Create (LPDIRECTDRAW2 lpDD, int x, int y, UINT cx, UINT cy, DWORD size,
						LPCTSTR szName)
{
	::EnterCriticalSection (&m_cs);

	DestroyHelper ();

	if (lpDD == NULL || cx  == 0 || cy == 0 || size == 0) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	m_dwSize = size;
	m_x = x;
	m_y = y;
	m_strName = szName;

	if (!m_imgImage.Create (lpDD, cx, cy)) {
		DestroyHelper ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

BOOL CDXSprite::LoadFile (LPDIRECTDRAW2 lpDD, LPCTSTR szBitmap, int x, int y, DWORD size,
						  LPCTSTR szName)
{
	::EnterCriticalSection (&m_cs);

	DestroyHelper ();

	if (lpDD == NULL || szBitmap == NULL || size == 0) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	m_dwSize = size;
	m_x = x;
	m_y = y;
	m_strName = szName;

	if (!m_imgImage.LoadFile (lpDD, szBitmap)) {
		DestroyHelper ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

BOOL CDXSprite::LoadResource (LPDIRECTDRAW2 lpDD, LPCTSTR szBitmap, int x, int y, DWORD size,
						LPCTSTR szName)
{
	::EnterCriticalSection (&m_cs);

	DestroyHelper ();

	if (lpDD == NULL || szBitmap == NULL || size == 0) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	m_dwSize = size;
	m_x = x;
	m_y = y;
	m_strName = szName;

	if (!m_imgImage.LoadResource (lpDD, szBitmap)) {
		DestroyHelper ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

void CDXSprite::Destroy (void)
{
	::EnterCriticalSection (&m_cs);
	DestroyHelper ();
	::LeaveCriticalSection (&m_cs);
}

void CDXSprite::DestroyHelper (void)
{
	m_imgImage.Destroy ();
	m_dwFrame = 0;
	m_dwSize = 0;
	m_x = 0;
	m_y = 0;
}

BOOL CDXSprite::FastDraw (LPDIRECTDRAWSURFACE3 lpDDDestSurface, DWORD dwFlags)
{
	::EnterCriticalSection (&m_cs);
	if (lpDDDestSurface == NULL || m_dwFrame >= m_dwSize) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	//   
	int pos = (m_dwFrame * m_imgImage.GetWidth ()) / m_dwSize;
	CRect rect (pos, 0, pos + m_imgImage.GetWidth () / m_dwSize, m_imgImage.GetHeight ());
	m_imgImage.FastDraw (lpDDDestSurface, rect, m_x, m_y, dwFlags);

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

BOOL CDXSprite::Draw (LPDIRECTDRAWSURFACE3 lpDDDestSurface, DWORD dwFlags, LPDDBLTFX lpDDBltFx)
{
	::EnterCriticalSection (&m_cs);
	if (lpDDDestSurface == NULL || m_dwFrame >= m_dwSize) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	//   
	int pos = (m_dwFrame * m_imgImage.GetWidth ()) / m_dwSize;
	CRect rect (pos, 0, pos + m_imgImage.GetWidth () / m_dwSize, m_imgImage.GetHeight ());
	m_imgImage.Draw (lpDDDestSurface, rect, m_x, m_y, dwFlags, lpDDBltFx);

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

void CDXSprite::SetFrame (DWORD dwNewFrame)
{
	::EnterCriticalSection (&m_cs);
	if (dwNewFrame < m_dwSize)
		m_dwFrame = dwNewFrame;
	::LeaveCriticalSection (&m_cs);
}

void CDXSprite::NextFrame (BOOL bLoop)
{
	::EnterCriticalSection (&m_cs);
	m_dwFrame++;
	if (m_dwFrame >= m_dwSize)
		if (bLoop)
			m_dwFrame = 0;
		else
			m_dwFrame--;
	::LeaveCriticalSection (&m_cs);
}

void CDXSprite::SetPos (int x, int y)
{
	::EnterCriticalSection (&m_cs);
	m_x = x;
	m_y = y;
	::LeaveCriticalSection (&m_cs);
}

int CDXSprite::GetWidth (void)
{
	::EnterCriticalSection (&m_cs);
	if (m_dwSize == 0) {
		::LeaveCriticalSection (&m_cs);
		return -1;
	}
	int n = m_imgImage.GetWidth () / m_dwSize;
	::LeaveCriticalSection (&m_cs);
	return n;
}

int CDXSprite::GetHeight (void)
{
	::EnterCriticalSection (&m_cs);
	int n = m_imgImage.GetHeight ();
	::LeaveCriticalSection (&m_cs);
	return n;
}

BOOL CDXSprite::IsLoaded (void)
{
	::EnterCriticalSection (&m_cs);
	BOOL ret = (m_imgImage.GetSurface3 () != NULL);
	::LeaveCriticalSection (&m_cs);
	return ret;
}

CString CDXSprite::GetName (void)
{
	::EnterCriticalSection (&m_cs);
	CString ret = m_strName;
	::LeaveCriticalSection (&m_cs);
	return ret;
}

DWORD CDXSprite::GetFrame (void)
{
	::EnterCriticalSection (&m_cs);
	DWORD ret = m_dwFrame;
	::LeaveCriticalSection (&m_cs);
	return ret;
}
